import pytest

from pybind11_tests import ConstructorStats

np = pytest.importorskip("numpy")
m = pytest.importorskip("pybind11_tests.eigen_matrix")


ref = np.array(
    [
        [0.0, 3, 0, 0, 0, 11],
        [22, 0, 0, 0, 17, 11],
        [7, 5, 0, 1, 0, 11],
        [0, 0, 0, 0, 0, 11],
        [0, 0, 14, 0, 8, 11],
    ]
)


def assert_equal_ref(mat):
    np.testing.assert_array_equal(mat, ref)


def assert_sparse_equal_ref(sparse_mat):
    assert_equal_ref(sparse_mat.toarray())


def test_fixed():
    assert_equal_ref(m.fixed_c())
    assert_equal_ref(m.fixed_r())
    assert_equal_ref(m.fixed_copy_r(m.fixed_r()))
    assert_equal_ref(m.fixed_copy_c(m.fixed_c()))
    assert_equal_ref(m.fixed_copy_r(m.fixed_c()))
    assert_equal_ref(m.fixed_copy_c(m.fixed_r()))


def test_dense():
    assert_equal_ref(m.dense_r())
    assert_equal_ref(m.dense_c())
    assert_equal_ref(m.dense_copy_r(m.dense_r()))
    assert_equal_ref(m.dense_copy_c(m.dense_c()))
    assert_equal_ref(m.dense_copy_r(m.dense_c()))
    assert_equal_ref(m.dense_copy_c(m.dense_r()))


def test_partially_fixed():
    ref2 = np.array([[0.0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14, 15]])
    np.testing.assert_array_equal(m.partial_copy_four_rm_r(ref2), ref2)
    np.testing.assert_array_equal(m.partial_copy_four_rm_c(ref2), ref2)
    np.testing.assert_array_equal(m.partial_copy_four_rm_r(ref2[:, 1]), ref2[:, [1]])
    np.testing.assert_array_equal(m.partial_copy_four_rm_c(ref2[0, :]), ref2[[0], :])
    np.testing.assert_array_equal(
        m.partial_copy_four_rm_r(ref2[:, (0, 2)]), ref2[:, (0, 2)]
    )
    np.testing.assert_array_equal(
        m.partial_copy_four_rm_c(ref2[(3, 1, 2), :]), ref2[(3, 1, 2), :]
    )

    np.testing.assert_array_equal(m.partial_copy_four_cm_r(ref2), ref2)
    np.testing.assert_array_equal(m.partial_copy_four_cm_c(ref2), ref2)
    np.testing.assert_array_equal(m.partial_copy_four_cm_r(ref2[:, 1]), ref2[:, [1]])
    np.testing.assert_array_equal(m.partial_copy_four_cm_c(ref2[0, :]), ref2[[0], :])
    np.testing.assert_array_equal(
        m.partial_copy_four_cm_r(ref2[:, (0, 2)]), ref2[:, (0, 2)]
    )
    np.testing.assert_array_equal(
        m.partial_copy_four_cm_c(ref2[(3, 1, 2), :]), ref2[(3, 1, 2), :]
    )

    # TypeError should be raise for a shape mismatch
    functions = [
        m.partial_copy_four_rm_r,
        m.partial_copy_four_rm_c,
        m.partial_copy_four_cm_r,
        m.partial_copy_four_cm_c,
    ]
    matrix_with_wrong_shape = [[1, 2], [3, 4]]
    for f in functions:
        with pytest.raises(TypeError) as excinfo:
            f(matrix_with_wrong_shape)
        assert "incompatible function arguments" in str(excinfo.value)


def test_mutator_descriptors():
    zr = np.arange(30, dtype="float32").reshape(5, 6)  # row-major
    zc = zr.reshape(6, 5).transpose()  # column-major

    m.fixed_mutator_r(zr)
    m.fixed_mutator_c(zc)
    m.fixed_mutator_a(zr)
    m.fixed_mutator_a(zc)
    with pytest.raises(TypeError) as excinfo:
        m.fixed_mutator_r(zc)
    assert (
        "(arg0: numpy.ndarray[numpy.float32[5, 6],"
        " flags.writeable, flags.c_contiguous]) -> None" in str(excinfo.value)
    )
    with pytest.raises(TypeError) as excinfo:
        m.fixed_mutator_c(zr)
    assert (
        "(arg0: numpy.ndarray[numpy.float32[5, 6],"
        " flags.writeable, flags.f_contiguous]) -> None" in str(excinfo.value)
    )
    with pytest.raises(TypeError) as excinfo:
        m.fixed_mutator_a(np.array([[1, 2], [3, 4]], dtype="float32"))
    assert "(arg0: numpy.ndarray[numpy.float32[5, 6], flags.writeable]) -> None" in str(
        excinfo.value
    )
    zr.flags.writeable = False
    with pytest.raises(TypeError):
        m.fixed_mutator_r(zr)
    with pytest.raises(TypeError):
        m.fixed_mutator_a(zr)


def test_cpp_casting():
    assert m.cpp_copy(m.fixed_r()) == 22.0
    assert m.cpp_copy(m.fixed_c()) == 22.0
    z = np.array([[5.0, 6], [7, 8]])
    assert m.cpp_copy(z) == 7.0
    assert m.cpp_copy(m.get_cm_ref()) == 21.0
    assert m.cpp_copy(m.get_rm_ref()) == 21.0
    assert m.cpp_ref_c(m.get_cm_ref()) == 21.0
    assert m.cpp_ref_r(m.get_rm_ref()) == 21.0
    with pytest.raises(RuntimeError) as excinfo:
        # Can't reference m.fixed_c: it contains floats, m.cpp_ref_any wants doubles
        m.cpp_ref_any(m.fixed_c())
    assert "Unable to cast Python instance" in str(excinfo.value)
    with pytest.raises(RuntimeError) as excinfo:
        # Can't reference m.fixed_r: it contains floats, m.cpp_ref_any wants doubles
        m.cpp_ref_any(m.fixed_r())
    assert "Unable to cast Python instance" in str(excinfo.value)
    assert m.cpp_ref_any(m.ReturnTester.create()) == 1.0

    assert m.cpp_ref_any(m.get_cm_ref()) == 21.0
    assert m.cpp_ref_any(m.get_cm_ref()) == 21.0


def test_pass_readonly_array():
    z = np.full((5, 6), 42.0)
    z.flags.writeable = False
    np.testing.assert_array_equal(z, m.fixed_copy_r(z))
    np.testing.assert_array_equal(m.fixed_r_const(), m.fixed_r())
    assert not m.fixed_r_const().flags.writeable
    np.testing.assert_array_equal(m.fixed_copy_r(m.fixed_r_const()), m.fixed_r_const())


def test_nonunit_stride_from_python():
    counting_mat = np.arange(9.0, dtype=np.float32).reshape((3, 3))
    second_row = counting_mat[1, :]
    second_col = counting_mat[:, 1]
    np.testing.assert_array_equal(m.double_row(second_row), 2.0 * second_row)
    np.testing.assert_array_equal(m.double_col(second_row), 2.0 * second_row)
    np.testing.assert_array_equal(m.double_complex(second_row), 2.0 * second_row)
    np.testing.assert_array_equal(m.double_row(second_col), 2.0 * second_col)
    np.testing.assert_array_equal(m.double_col(second_col), 2.0 * second_col)
    np.testing.assert_array_equal(m.double_complex(second_col), 2.0 * second_col)

    counting_3d = np.arange(27.0, dtype=np.float32).reshape((3, 3, 3))
    slices = [counting_3d[0, :, :], counting_3d[:, 0, :], counting_3d[:, :, 0]]
    for ref_mat in slices:
        np.testing.assert_array_equal(m.double_mat_cm(ref_mat), 2.0 * ref_mat)
        np.testing.assert_array_equal(m.double_mat_rm(ref_mat), 2.0 * ref_mat)

    # Mutator:
    m.double_threer(second_row)
    m.double_threec(second_col)
    np.testing.assert_array_equal(counting_mat, [[0.0, 2, 2], [6, 16, 10], [6, 14, 8]])


def test_negative_stride_from_python(msg):
    """Eigen doesn't support (as of yet) negative strides. When a function takes an Eigen matrix by
    copy or const reference, we can pass a numpy array that has negative strides.  Otherwise, an
    exception will be thrown as Eigen will not be able to map the numpy array."""

    counting_mat = np.arange(9.0, dtype=np.float32).reshape((3, 3))
    counting_mat = counting_mat[::-1, ::-1]
    second_row = counting_mat[1, :]
    second_col = counting_mat[:, 1]
    np.testing.assert_array_equal(m.double_row(second_row), 2.0 * second_row)
    np.testing.assert_array_equal(m.double_col(second_row), 2.0 * second_row)
    np.testing.assert_array_equal(m.double_complex(second_row), 2.0 * second_row)
    np.testing.assert_array_equal(m.double_row(second_col), 2.0 * second_col)
    np.testing.assert_array_equal(m.double_col(second_col), 2.0 * second_col)
    np.testing.assert_array_equal(m.double_complex(second_col), 2.0 * second_col)

    counting_3d = np.arange(27.0, dtype=np.float32).reshape((3, 3, 3))
    counting_3d = counting_3d[::-1, ::-1, ::-1]
    slices = [counting_3d[0, :, :], counting_3d[:, 0, :], counting_3d[:, :, 0]]
    for ref_mat in slices:
        np.testing.assert_array_equal(m.double_mat_cm(ref_mat), 2.0 * ref_mat)
        np.testing.assert_array_equal(m.double_mat_rm(ref_mat), 2.0 * ref_mat)

    # Mutator:
    with pytest.raises(TypeError) as excinfo:
        m.double_threer(second_row)
    assert (
        msg(excinfo.value)
        == """
        double_threer(): incompatible function arguments. The following argument types are supported:
            1. (arg0: numpy.ndarray[numpy.float32[1, 3], flags.writeable]) -> None

        Invoked with: """
        + repr(np.array([5.0, 4.0, 3.0], dtype="float32"))
    )

    with pytest.raises(TypeError) as excinfo:
        m.double_threec(second_col)
    assert (
        msg(excinfo.value)
        == """
        double_threec(): incompatible function arguments. The following argument types are supported:
            1. (arg0: numpy.ndarray[numpy.float32[3, 1], flags.writeable]) -> None

        Invoked with: """
        + repr(np.array([7.0, 4.0, 1.0], dtype="float32"))
    )


def test_block_runtime_error_type_caster_eigen_ref_made_a_copy():
    with pytest.raises(RuntimeError) as excinfo:
        m.block(ref, 0, 0, 0, 0)
    assert str(excinfo.value) == "type_caster for Eigen::Ref made a copy."


def test_nonunit_stride_to_python():
    assert np.all(m.diagonal(ref) == ref.diagonal())
    assert np.all(m.diagonal_1(ref) == ref.diagonal(1))
    for i in range(-5, 7):
        assert np.all(m.diagonal_n(ref, i) == ref.diagonal(i)), f"m.diagonal_n({i})"

    # Must be order="F", otherwise the type_caster will make a copy and
    # m.block() will return a dangling reference (heap-use-after-free).
    rof = np.asarray(ref, order="F")
    assert np.all(m.block(rof, 2, 1, 3, 3) == rof[2:5, 1:4])
    assert np.all(m.block(rof, 1, 4, 4, 2) == rof[1:, 4:])
    assert np.all(m.block(rof, 1, 4, 3, 2) == rof[1:4, 4:])


def test_eigen_ref_to_python():
    chols = [m.cholesky1, m.cholesky2, m.cholesky3, m.cholesky4]
    for i, chol in enumerate(chols, start=1):
        mymat = chol(np.array([[1.0, 2, 4], [2, 13, 23], [4, 23, 77]]))
        assert np.all(
            mymat == np.array([[1, 0, 0], [2, 3, 0], [4, 5, 6]])
        ), f"cholesky{i}"


def assign_both(a1, a2, r, c, v):
    a1[r, c] = v
    a2[r, c] = v


def array_copy_but_one(a, r, c, v):
    z = np.array(a, copy=True)
    z[r, c] = v
    return z


def test_eigen_return_references():
    """Tests various ways of returning references and non-referencing copies"""

    primary = np.ones((10, 10))
    a = m.ReturnTester()
    a_get1 = a.get()
    assert not a_get1.flags.owndata
    assert a_get1.flags.writeable
    assign_both(a_get1, primary, 3, 3, 5)
    a_get2 = a.get_ptr()
    assert not a_get2.flags.owndata
    assert a_get2.flags.writeable
    assign_both(a_get1, primary, 2, 3, 6)

    a_view1 = a.view()
    assert not a_view1.flags.owndata
    assert not a_view1.flags.writeable
    with pytest.raises(ValueError):
        a_view1[2, 3] = 4
    a_view2 = a.view_ptr()
    assert not a_view2.flags.owndata
    assert not a_view2.flags.writeable
    with pytest.raises(ValueError):
        a_view2[2, 3] = 4

    a_copy1 = a.copy_get()
    assert a_copy1.flags.owndata
    assert a_copy1.flags.writeable
    np.testing.assert_array_equal(a_copy1, primary)
    a_copy1[7, 7] = -44  # Shouldn't affect anything else
    c1want = array_copy_but_one(primary, 7, 7, -44)
    a_copy2 = a.copy_view()
    assert a_copy2.flags.owndata
    assert a_copy2.flags.writeable
    np.testing.assert_array_equal(a_copy2, primary)
    a_copy2[4, 4] = -22  # Shouldn't affect anything else
    c2want = array_copy_but_one(primary, 4, 4, -22)

    a_ref1 = a.ref()
    assert not a_ref1.flags.owndata
    assert a_ref1.flags.writeable
    assign_both(a_ref1, primary, 1, 1, 15)
    a_ref2 = a.ref_const()
    assert not a_ref2.flags.owndata
    assert not a_ref2.flags.writeable
    with pytest.raises(ValueError):
        a_ref2[5, 5] = 33
    a_ref3 = a.ref_safe()
    assert not a_ref3.flags.owndata
    assert a_ref3.flags.writeable
    assign_both(a_ref3, primary, 0, 7, 99)
    a_ref4 = a.ref_const_safe()
    assert not a_ref4.flags.owndata
    assert not a_ref4.flags.writeable
    with pytest.raises(ValueError):
        a_ref4[7, 0] = 987654321

    a_copy3 = a.copy_ref()
    assert a_copy3.flags.owndata
    assert a_copy3.flags.writeable
    np.testing.assert_array_equal(a_copy3, primary)
    a_copy3[8, 1] = 11
    c3want = array_copy_but_one(primary, 8, 1, 11)
    a_copy4 = a.copy_ref_const()
    assert a_copy4.flags.owndata
    assert a_copy4.flags.writeable
    np.testing.assert_array_equal(a_copy4, primary)
    a_copy4[8, 4] = 88
    c4want = array_copy_but_one(primary, 8, 4, 88)

    a_block1 = a.block(3, 3, 2, 2)
    assert not a_block1.flags.owndata
    assert a_block1.flags.writeable
    a_block1[0, 0] = 55
    primary[3, 3] = 55
    a_block2 = a.block_safe(2, 2, 3, 2)
    assert not a_block2.flags.owndata
    assert a_block2.flags.writeable
    a_block2[2, 1] = -123
    primary[4, 3] = -123
    a_block3 = a.block_const(6, 7, 4, 3)
    assert not a_block3.flags.owndata
    assert not a_block3.flags.writeable
    with pytest.raises(ValueError):
        a_block3[2, 2] = -44444

    a_copy5 = a.copy_block(2, 2, 2, 3)
    assert a_copy5.flags.owndata
    assert a_copy5.flags.writeable
    np.testing.assert_array_equal(a_copy5, primary[2:4, 2:5])
    a_copy5[1, 1] = 777
    c5want = array_copy_but_one(primary[2:4, 2:5], 1, 1, 777)

    a_corn1 = a.corners()
    assert not a_corn1.flags.owndata
    assert a_corn1.flags.writeable
    a_corn1 *= 50
    a_corn1[1, 1] = 999
    primary[0, 0] = 50
    primary[0, 9] = 50
    primary[9, 0] = 50
    primary[9, 9] = 999
    a_corn2 = a.corners_const()
    assert not a_corn2.flags.owndata
    assert not a_corn2.flags.writeable
    with pytest.raises(ValueError):
        a_corn2[1, 0] = 51

    # All of the changes made all the way along should be visible everywhere
    # now (except for the copies, of course)
    np.testing.assert_array_equal(a_get1, primary)
    np.testing.assert_array_equal(a_get2, primary)
    np.testing.assert_array_equal(a_view1, primary)
    np.testing.assert_array_equal(a_view2, primary)
    np.testing.assert_array_equal(a_ref1, primary)
    np.testing.assert_array_equal(a_ref2, primary)
    np.testing.assert_array_equal(a_ref3, primary)
    np.testing.assert_array_equal(a_ref4, primary)
    np.testing.assert_array_equal(a_block1, primary[3:5, 3:5])
    np.testing.assert_array_equal(a_block2, primary[2:5, 2:4])
    np.testing.assert_array_equal(a_block3, primary[6:10, 7:10])
    np.testing.assert_array_equal(
        a_corn1, primary[0 :: primary.shape[0] - 1, 0 :: primary.shape[1] - 1]
    )
    np.testing.assert_array_equal(
        a_corn2, primary[0 :: primary.shape[0] - 1, 0 :: primary.shape[1] - 1]
    )

    np.testing.assert_array_equal(a_copy1, c1want)
    np.testing.assert_array_equal(a_copy2, c2want)
    np.testing.assert_array_equal(a_copy3, c3want)
    np.testing.assert_array_equal(a_copy4, c4want)
    np.testing.assert_array_equal(a_copy5, c5want)


def assert_keeps_alive(cl, method, *args):
    cstats = ConstructorStats.get(cl)
    start_with = cstats.alive()
    a = cl()
    assert cstats.alive() == start_with + 1
    z = method(a, *args)
    assert cstats.alive() == start_with + 1
    del a
    # Here's the keep alive in action:
    assert cstats.alive() == start_with + 1
    del z
    # Keep alive should have expired:
    assert cstats.alive() == start_with


def test_eigen_keepalive():
    a = m.ReturnTester()
    cstats = ConstructorStats.get(m.ReturnTester)
    assert cstats.alive() == 1
    unsafe = [a.ref(), a.ref_const(), a.block(1, 2, 3, 4)]
    copies = [
        a.copy_get(),
        a.copy_view(),
        a.copy_ref(),
        a.copy_ref_const(),
        a.copy_block(4, 3, 2, 1),
    ]
    del a
    assert cstats.alive() == 0
    del unsafe
    del copies

    for meth in [
        m.ReturnTester.get,
        m.ReturnTester.get_ptr,
        m.ReturnTester.view,
        m.ReturnTester.view_ptr,
        m.ReturnTester.ref_safe,
        m.ReturnTester.ref_const_safe,
        m.ReturnTester.corners,
        m.ReturnTester.corners_const,
    ]:
        assert_keeps_alive(m.ReturnTester, meth)

    for meth in [m.ReturnTester.block_safe, m.ReturnTester.block_const]:
        assert_keeps_alive(m.ReturnTester, meth, 4, 3, 2, 1)


def test_eigen_ref_mutators():
    """Tests Eigen's ability to mutate numpy values"""

    orig = np.array([[1.0, 2, 3], [4, 5, 6], [7, 8, 9]])
    zr = np.array(orig)
    zc = np.array(orig, order="F")
    m.add_rm(zr, 1, 0, 100)
    assert np.all(zr == np.array([[1.0, 2, 3], [104, 5, 6], [7, 8, 9]]))
    m.add_cm(zc, 1, 0, 200)
    assert np.all(zc == np.array([[1.0, 2, 3], [204, 5, 6], [7, 8, 9]]))

    m.add_any(zr, 1, 0, 20)
    assert np.all(zr == np.array([[1.0, 2, 3], [124, 5, 6], [7, 8, 9]]))
    m.add_any(zc, 1, 0, 10)
    assert np.all(zc == np.array([[1.0, 2, 3], [214, 5, 6], [7, 8, 9]]))

    # Can't reference a col-major array with a row-major Ref, and vice versa:
    with pytest.raises(TypeError):
        m.add_rm(zc, 1, 0, 1)
    with pytest.raises(TypeError):
        m.add_cm(zr, 1, 0, 1)

    # Overloads:
    m.add1(zr, 1, 0, -100)
    m.add2(zr, 1, 0, -20)
    assert np.all(zr == orig)
    m.add1(zc, 1, 0, -200)
    m.add2(zc, 1, 0, -10)
    assert np.all(zc == orig)

    # a non-contiguous slice (this won't work on either the row- or
    # column-contiguous refs, but should work for the any)
    cornersr = zr[0::2, 0::2]
    cornersc = zc[0::2, 0::2]

    assert np.all(cornersr == np.array([[1.0, 3], [7, 9]]))
    assert np.all(cornersc == np.array([[1.0, 3], [7, 9]]))

    with pytest.raises(TypeError):
        m.add_rm(cornersr, 0, 1, 25)
    with pytest.raises(TypeError):
        m.add_cm(cornersr, 0, 1, 25)
    with pytest.raises(TypeError):
        m.add_rm(cornersc, 0, 1, 25)
    with pytest.raises(TypeError):
        m.add_cm(cornersc, 0, 1, 25)
    m.add_any(cornersr, 0, 1, 25)
    m.add_any(cornersc, 0, 1, 44)
    assert np.all(zr == np.array([[1.0, 2, 28], [4, 5, 6], [7, 8, 9]]))
    assert np.all(zc == np.array([[1.0, 2, 47], [4, 5, 6], [7, 8, 9]]))

    # You shouldn't be allowed to pass a non-writeable array to a mutating Eigen method:
    zro = zr[0:4, 0:4]
    zro.flags.writeable = False
    with pytest.raises(TypeError):
        m.add_rm(zro, 0, 0, 0)
    with pytest.raises(TypeError):
        m.add_any(zro, 0, 0, 0)
    with pytest.raises(TypeError):
        m.add1(zro, 0, 0, 0)
    with pytest.raises(TypeError):
        m.add2(zro, 0, 0, 0)

    # integer array shouldn't be passable to a double-matrix-accepting mutating func:
    zi = np.array([[1, 2], [3, 4]])
    with pytest.raises(TypeError):
        m.add_rm(zi)


def test_numpy_ref_mutators():
    """Tests numpy mutating Eigen matrices (for returned Eigen::Ref<...>s)"""

    m.reset_refs()  # In case another test already changed it

    zc = m.get_cm_ref()
    zcro = m.get_cm_const_ref()
    zr = m.get_rm_ref()
    zrro = m.get_rm_const_ref()

    assert [zc[1, 2], zcro[1, 2], zr[1, 2], zrro[1, 2]] == [23] * 4

    assert not zc.flags.owndata
    assert zc.flags.writeable
    assert not zr.flags.owndata
    assert zr.flags.writeable
    assert not zcro.flags.owndata
    assert not zcro.flags.writeable
    assert not zrro.flags.owndata
    assert not zrro.flags.writeable

    zc[1, 2] = 99
    expect = np.array([[11.0, 12, 13], [21, 22, 99], [31, 32, 33]])
    # We should have just changed zc, of course, but also zcro and the original eigen matrix
    assert np.all(zc == expect)
    assert np.all(zcro == expect)
    assert np.all(m.get_cm_ref() == expect)

    zr[1, 2] = 99
    assert np.all(zr == expect)
    assert np.all(zrro == expect)
    assert np.all(m.get_rm_ref() == expect)

    # Make sure the readonly ones are numpy-readonly:
    with pytest.raises(ValueError):
        zcro[1, 2] = 6
    with pytest.raises(ValueError):
        zrro[1, 2] = 6

    # We should be able to explicitly copy like this (and since we're copying,
    # the const should drop away)
    y1 = np.array(m.get_cm_const_ref())

    assert y1.flags.owndata
    assert y1.flags.writeable
    # We should get copies of the eigen data, which was modified above:
    assert y1[1, 2] == 99
    y1[1, 2] += 12
    assert y1[1, 2] == 111
    assert zc[1, 2] == 99  # Make sure we aren't referencing the original


def test_both_ref_mutators():
    """Tests a complex chain of nested eigen/numpy references"""

    m.reset_refs()  # In case another test already changed it

    z = m.get_cm_ref()  # numpy -> eigen
    z[0, 2] -= 3
    z2 = m.incr_matrix(z, 1)  # numpy -> eigen -> numpy -> eigen
    z2[1, 1] += 6
    z3 = m.incr_matrix(z, 2)  # (numpy -> eigen)^3
    z3[2, 2] += -5
    z4 = m.incr_matrix(z, 3)  # (numpy -> eigen)^4
    z4[1, 1] -= 1
    z5 = m.incr_matrix(z, 4)  # (numpy -> eigen)^5
    z5[0, 0] = 0
    assert np.all(z == z2)
    assert np.all(z == z3)
    assert np.all(z == z4)
    assert np.all(z == z5)
    expect = np.array([[0.0, 22, 20], [31, 37, 33], [41, 42, 38]])
    assert np.all(z == expect)

    y = np.array(range(100), dtype="float64").reshape(10, 10)
    y2 = m.incr_matrix_any(y, 10)  # np -> eigen -> np
    y3 = m.incr_matrix_any(
        y2[0::2, 0::2], -33
    )  # np -> eigen -> np slice -> np -> eigen -> np
    y4 = m.even_rows(y3)  # numpy -> eigen slice -> (... y3)
    y5 = m.even_cols(y4)  # numpy -> eigen slice -> (... y4)
    y6 = m.incr_matrix_any(y5, 1000)  # numpy -> eigen -> (... y5)

    # Apply same mutations using just numpy:
    yexpect = np.array(range(100), dtype="float64").reshape(10, 10)
    yexpect += 10
    yexpect[0::2, 0::2] -= 33
    yexpect[0::4, 0::4] += 1000
    assert np.all(y6 == yexpect[0::4, 0::4])
    assert np.all(y5 == yexpect[0::4, 0::4])
    assert np.all(y4 == yexpect[0::4, 0::2])
    assert np.all(y3 == yexpect[0::2, 0::2])
    assert np.all(y2 == yexpect)
    assert np.all(y == yexpect)


def test_nocopy_wrapper():
    # get_elem requires a column-contiguous matrix reference, but should be
    # callable with other types of matrix (via copying):
    int_matrix_colmajor = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], order="F")
    dbl_matrix_colmajor = np.array(
        int_matrix_colmajor, dtype="double", order="F", copy=True
    )
    int_matrix_rowmajor = np.array(int_matrix_colmajor, order="C", copy=True)
    dbl_matrix_rowmajor = np.array(
        int_matrix_rowmajor, dtype="double", order="C", copy=True
    )

    # All should be callable via get_elem:
    assert m.get_elem(int_matrix_colmajor) == 8
    assert m.get_elem(dbl_matrix_colmajor) == 8
    assert m.get_elem(int_matrix_rowmajor) == 8
    assert m.get_elem(dbl_matrix_rowmajor) == 8

    # All but the second should fail with m.get_elem_nocopy:
    with pytest.raises(TypeError) as excinfo:
        m.get_elem_nocopy(int_matrix_colmajor)
    assert "get_elem_nocopy(): incompatible function arguments." in str(excinfo.value)
    assert ", flags.f_contiguous" in str(excinfo.value)
    assert m.get_elem_nocopy(dbl_matrix_colmajor) == 8
    with pytest.raises(TypeError) as excinfo:
        m.get_elem_nocopy(int_matrix_rowmajor)
    assert "get_elem_nocopy(): incompatible function arguments." in str(excinfo.value)
    assert ", flags.f_contiguous" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.get_elem_nocopy(dbl_matrix_rowmajor)
    assert "get_elem_nocopy(): incompatible function arguments." in str(excinfo.value)
    assert ", flags.f_contiguous" in str(excinfo.value)

    # For the row-major test, we take a long matrix in row-major, so only the third is allowed:
    with pytest.raises(TypeError) as excinfo:
        m.get_elem_rm_nocopy(int_matrix_colmajor)
    assert "get_elem_rm_nocopy(): incompatible function arguments." in str(
        excinfo.value
    )
    assert ", flags.c_contiguous" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.get_elem_rm_nocopy(dbl_matrix_colmajor)
    assert "get_elem_rm_nocopy(): incompatible function arguments." in str(
        excinfo.value
    )
    assert ", flags.c_contiguous" in str(excinfo.value)
    assert m.get_elem_rm_nocopy(int_matrix_rowmajor) == 8
    with pytest.raises(TypeError) as excinfo:
        m.get_elem_rm_nocopy(dbl_matrix_rowmajor)
    assert "get_elem_rm_nocopy(): incompatible function arguments." in str(
        excinfo.value
    )
    assert ", flags.c_contiguous" in str(excinfo.value)


def test_eigen_ref_life_support():
    """Ensure the lifetime of temporary arrays created by the `Ref` caster

    The `Ref` caster sometimes creates a copy which needs to stay alive. This needs to
    happen both for directs casts (just the array) or indirectly (e.g. list of arrays).
    """

    a = np.full(shape=10, fill_value=8, dtype=np.int8)
    assert m.get_elem_direct(a) == 8

    list_of_a = [a]
    assert m.get_elem_indirect(list_of_a) == 8


def test_special_matrix_objects():
    assert np.all(m.incr_diag(7) == np.diag([1.0, 2, 3, 4, 5, 6, 7]))

    asymm = np.array([[1.0, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    symm_lower = np.array(asymm)
    symm_upper = np.array(asymm)
    for i in range(4):
        for j in range(i + 1, 4):
            symm_lower[i, j] = symm_lower[j, i]
            symm_upper[j, i] = symm_upper[i, j]

    assert np.all(m.symmetric_lower(asymm) == symm_lower)
    assert np.all(m.symmetric_upper(asymm) == symm_upper)


def test_dense_signature(doc):
    assert (
        doc(m.double_col)
        == """
        double_col(arg0: numpy.ndarray[numpy.float32[m, 1]]) -> numpy.ndarray[numpy.float32[m, 1]]
    """
    )
    assert (
        doc(m.double_row)
        == """
        double_row(arg0: numpy.ndarray[numpy.float32[1, n]]) -> numpy.ndarray[numpy.float32[1, n]]
    """
    )
    assert doc(m.double_complex) == (
        """
        double_complex(arg0: numpy.ndarray[numpy.complex64[m, 1]])"""
        """ -> numpy.ndarray[numpy.complex64[m, 1]]
    """
    )
    assert doc(m.double_mat_rm) == (
        """
        double_mat_rm(arg0: numpy.ndarray[numpy.float32[m, n]])"""
        """ -> numpy.ndarray[numpy.float32[m, n]]
    """
    )


def test_defaults(doc):
    assert "\n" not in str(doc(m.defaults_mat))
    assert "\n" not in str(doc(m.defaults_vec))


def test_named_arguments():
    a = np.array([[1.0, 2], [3, 4], [5, 6]])
    b = np.ones((2, 1))

    assert np.all(m.matrix_multiply(a, b) == np.array([[3.0], [7], [11]]))
    assert np.all(m.matrix_multiply(A=a, B=b) == np.array([[3.0], [7], [11]]))
    assert np.all(m.matrix_multiply(B=b, A=a) == np.array([[3.0], [7], [11]]))

    with pytest.raises(ValueError) as excinfo:
        m.matrix_multiply(b, a)
    assert str(excinfo.value) == "Nonconformable matrices!"

    with pytest.raises(ValueError) as excinfo:
        m.matrix_multiply(A=b, B=a)
    assert str(excinfo.value) == "Nonconformable matrices!"

    with pytest.raises(ValueError) as excinfo:
        m.matrix_multiply(B=a, A=b)
    assert str(excinfo.value) == "Nonconformable matrices!"


def test_sparse():
    pytest.importorskip("scipy")
    assert_sparse_equal_ref(m.sparse_r())
    assert_sparse_equal_ref(m.sparse_c())
    assert_sparse_equal_ref(m.sparse_copy_r(m.sparse_r()))
    assert_sparse_equal_ref(m.sparse_copy_c(m.sparse_c()))
    assert_sparse_equal_ref(m.sparse_copy_r(m.sparse_c()))
    assert_sparse_equal_ref(m.sparse_copy_c(m.sparse_r()))


def test_sparse_signature(doc):
    pytest.importorskip("scipy")
    assert (
        doc(m.sparse_copy_r)
        == """
        sparse_copy_r(arg0: scipy.sparse.csr_matrix[numpy.float32]) -> scipy.sparse.csr_matrix[numpy.float32]
    """
    )
    assert (
        doc(m.sparse_copy_c)
        == """
        sparse_copy_c(arg0: scipy.sparse.csc_matrix[numpy.float32]) -> scipy.sparse.csc_matrix[numpy.float32]
    """
    )


def test_issue738():
    """Ignore strides on a length-1 dimension (even if they would be incompatible length > 1)"""
    assert np.all(m.iss738_f1(np.array([[1.0, 2, 3]])) == np.array([[1.0, 102, 203]]))
    assert np.all(
        m.iss738_f1(np.array([[1.0], [2], [3]])) == np.array([[1.0], [12], [23]])
    )

    assert np.all(m.iss738_f2(np.array([[1.0, 2, 3]])) == np.array([[1.0, 102, 203]]))
    assert np.all(
        m.iss738_f2(np.array([[1.0], [2], [3]])) == np.array([[1.0], [12], [23]])
    )


@pytest.mark.parametrize("func", [m.iss738_f1, m.iss738_f2])
@pytest.mark.parametrize("sizes", [(0, 2), (2, 0)])
def test_zero_length(func, sizes):
    """Ignore strides on a length-0 dimension (even if they would be incompatible length > 1)"""
    assert np.all(func(np.zeros(sizes)) == np.zeros(sizes))


def test_issue1105():
    """Issue 1105: 1xN or Nx1 input arrays weren't accepted for eigen
    compile-time row vectors or column vector"""
    assert m.iss1105_row(np.ones((1, 7)))
    assert m.iss1105_col(np.ones((7, 1)))

    # These should still fail (incompatible dimensions):
    with pytest.raises(TypeError) as excinfo:
        m.iss1105_row(np.ones((7, 1)))
    assert "incompatible function arguments" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.iss1105_col(np.ones((1, 7)))
    assert "incompatible function arguments" in str(excinfo.value)


def test_custom_operator_new():
    """Using Eigen types as member variables requires a class-specific
    operator new with proper alignment"""

    o = m.CustomOperatorNew()
    np.testing.assert_allclose(o.a, 0.0)
    np.testing.assert_allclose(o.b.diagonal(), 1.0)



"""Copyright 2023 Simeon Ehrig

This file is part of alpaka.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.


Generate GitLab-CI test jobs yaml for the vikunja CI."""

import argparse
import sys, os, random
from typing import List, Dict, Tuple
from collections import OrderedDict

import alpaka_job_coverage as ajc
from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.util import filter_job_list, reorder_job_list

from versions import (
    get_sw_tuple_list,
    get_compiler_versions,
    get_backend_matrix,
)
from alpaka_filter import alpaka_post_filter
from custom_job import add_custom_jobs
from reorder_jobs import reorder_jobs
from generate_job_yaml import (
    generate_job_yaml_list,
    write_job_yaml,
    distribute_to_waves,
    JOB_COMPILE_ONLY,
    JOB_RUNTIME,
    JOB_UNKNOWN,
    WAVE_GROUP_NAMES,
)
from job_modifier import add_job_parameters
from verify import verify


def get_args() -> argparse.Namespace:
    """Define and parse the commandline arguments.

    Returns:
        argparse.Namespace: The commandline arguments.
    """
    parser = argparse.ArgumentParser(
        description="Calculate job matrix and create GitLab CI .yml."
    )

    parser.add_argument(
        "version", type=float, help="Version number of the used CI container."
    )
    parser.add_argument(
        "--print-combinations",
        action="store_true",
        help="Display combination matrix.",
    )
    parser.add_argument(
        "--verify", action="store_true", help="Verify generated combination matrix"
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="Combine flags: --print-combinations and --verify",
    )

    parser.add_argument(
        "-o",
        "--output-path",
        type=str,
        default="./jobs.yml",
        help="Path of the generated jobs yaml.",
    )

    parser.add_argument(
        "--filter",
        type=str,
        default="",
        help="Filter the jobs with a Python regex that checks the job names.",
    )

    parser.add_argument(
        "--reorder",
        type=str,
        default="",
        help="Orders jobs by their names. Expects a string consisting of one or more Python regex. "
        'The regex are separated by whitespaces. For example, the regex "^NVCC ^GCC" has the '
        "behavior that all NVCC jobs are executed first and then all GCC jobs.",
    )

    parser.add_argument(
        "--compile-only", action="store_true", help="Generate only compile only jobs."
    )

    parser.add_argument(
        "--runtime-only", action="store_true", help="Generate only runtime jobs."
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = get_args()

    # setup the parameters
    parameters: OrderedDict = OrderedDict()
    enable_clang_cuda = True
    parameters[HOST_COMPILER] = get_compiler_versions(clang_cuda=enable_clang_cuda)
    parameters[DEVICE_COMPILER] = get_compiler_versions(clang_cuda=enable_clang_cuda)
    # TODO(SimeonEhrig): remove GCC and Clang from DEVICE_COMPILER to disable CPU accelerator
    # Backends
    parameters[DEVICE_COMPILER] = list(
        filter(
            lambda compiler: compiler[NAME] != GCC and compiler[NAME] != CLANG,
            parameters[DEVICE_COMPILER],
        )
    )

    # TODO(SimeonEhrig): remove Clang 15 as CUDA compiler until issue #1906 is solved
    parameters[HOST_COMPILER].remove((CLANG_CUDA, "15"))
    parameters[DEVICE_COMPILER].remove((CLANG_CUDA, "15"))

    parameters[BACKENDS] = get_backend_matrix()
    parameters[CMAKE] = get_sw_tuple_list(CMAKE)
    parameters[BOOST] = get_sw_tuple_list(BOOST)
    parameters[UBUNTU] = get_sw_tuple_list(UBUNTU)
    parameters[CXX_STANDARD] = get_sw_tuple_list(CXX_STANDARD)
    parameters[BUILD_TYPE] = get_sw_tuple_list(BUILD_TYPE)
    parameters[JOB_EXECUTION_TYPE] = get_sw_tuple_list(JOB_EXECUTION_TYPE)

    job_matrix: List[Dict[str, Tuple[str, str]]] = ajc.create_job_list(
        parameters=parameters,
        post_filter=alpaka_post_filter,
        pair_size=2,
    )

    add_job_parameters(job_matrix)

    if args.print_combinations or args.all:
        print(f"number of combinations before reorder: {len(job_matrix)}")

    ajc.shuffle_job_matrix(job_matrix)
    reorder_jobs(job_matrix)

    if args.print_combinations or args.all:
        for compiler in job_matrix:
            print(compiler)

        print(f"number of combinations: {len(job_matrix)}")

    if args.verify or args.all:
        if not verify(job_matrix):
            sys.exit(1)

    job_matrix_yaml = generate_job_yaml_list(
        job_matrix=job_matrix, container_version=args.version
    )

    add_custom_jobs(job_matrix_yaml, args.version)
    # shuffle jobs to better utilize the special runner
    random.Random(42).shuffle(job_matrix_yaml)

    filter_regix = args.filter
    reorder_regix = args.reorder

    COMMIT_MESSAGE_FILTER_PREFIX = "CI_FILTER:"
    COMMIT_MESSAGE_REORDER_PREFIX = "CI_REORDER:"

    # If the environment variable CI_COMMIT_MESSAGE exists (like in GitLabCI Job),
    # check for the prefixes and overwrite argument values of --filter and --reorder if
    # prefix was found in the beginning of a line.
    if os.getenv("CI_COMMIT_MESSAGE"):
        for line in os.getenv("CI_COMMIT_MESSAGE").split("\n"):
            striped_line = line.strip()
            if striped_line.strip().startswith(COMMIT_MESSAGE_FILTER_PREFIX):
                filter_regix = striped_line[len(COMMIT_MESSAGE_FILTER_PREFIX) :].strip()
            if striped_line.startswith(COMMIT_MESSAGE_REORDER_PREFIX):
                reorder_regix = striped_line[
                    len(COMMIT_MESSAGE_REORDER_PREFIX) :
                ].strip()

    if filter_regix:
        job_matrix_yaml = filter_job_list(job_matrix_yaml, filter_regix)

    if reorder_regix:
        job_matrix_yaml = reorder_job_list(job_matrix_yaml, reorder_regix)

    # wave_job_matrix = ajc.distribute_to_waves(job_matrix_yaml, 10)
    wave_job_matrix = distribute_to_waves(job_matrix_yaml, {JOB_COMPILE_ONLY: 20})

    if wave_job_matrix[JOB_UNKNOWN]:
        print('\033[33mWARNING: Generator distribute jobs of type "JOB_UNKNOWN"\033[m')
        for wave in wave_job_matrix[JOB_UNKNOWN]:
            for job in wave:
                print(job)

    if args.compile_only:
        wave_job_matrix = {JOB_COMPILE_ONLY: wave_job_matrix[JOB_COMPILE_ONLY]}
        for wave_name in WAVE_GROUP_NAMES:
            if not wave_name in wave_job_matrix:
                wave_job_matrix[wave_name] = []

    if args.runtime_only:
        wave_job_matrix = {JOB_RUNTIME: wave_job_matrix[JOB_RUNTIME]}
        for wave_name in WAVE_GROUP_NAMES:
            if not wave_name in wave_job_matrix:
                wave_job_matrix[wave_name] = []

    write_job_yaml(
        job_matrix=wave_job_matrix,
        path=args.output_path,
    )



"""Copyright 2023 Simeon Ehrig

This file is part of alpaka.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.


Used software in the CI tests."""

from typing import Dict, List, Tuple
from typeguard import typechecked

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_globals import *  # pylint: disable=wildcard-import,unused-wildcard-import


# TODO: only an example
sw_versions: Dict[str, List[str]] = {
    GCC: ["9", "10", "11", "12"],
    CLANG: ["9", "10", "11", "12", "13", "14", "15"],
    NVCC: [
        "11.0",
        "11.1",
        "11.2",
        "11.3",
        "11.4",
        "11.5",
        "11.6",
        "11.7",
        "11.8",
        "12.0",
    ],
    HIPCC: ["5.0", "5.1", "5.2", "5.3"],
    BACKENDS: [
        #         ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE,
        ALPAKA_ACC_GPU_CUDA_ENABLE,
        ALPAKA_ACC_GPU_HIP_ENABLE,
    ],
    UBUNTU: ["20.04"],
    CMAKE: ["3.18.6", "3.19.8", "3.20.6", "3.21.6", "3.22.3", "3.23.2"],
    BOOST: ["1.74.0", "1.75.0", "1.76.0", "1.77.0", "1.78.0", "1.79.0", "1.80.0"],
    CXX_STANDARD: ["17", "20"],
    BUILD_TYPE: BUILD_TYPES,
    # use only TEST_COMPILE_ONLY, because TEST_RUNTIME will be set manually depend on some
    # conditions later
    JOB_EXECUTION_TYPE: [JOB_EXECUTION_COMPILE_ONLY],
}


@typechecked
def get_compiler_versions(clang_cuda: bool = True) -> List[Tuple[str, str]]:
    """Generate a list of compiler name version tuple.

    Args:
        clang_cuda (bool, optional): If true, create entries for cling-cuda basing on the clang
        version. Defaults to True.

    Returns:
        List[Tuple[str, str]]: The compiler name version tuple list.
    """
    compilers: List[Tuple[str, str]] = []

    # only use keys defined in sw_versions
    for compiler_name in set(sw_versions.keys()).intersection(
        [GCC, CLANG, NVCC, HIPCC]
    ):
        for version in sw_versions[compiler_name]:
            compilers.append((compiler_name, version))
            if clang_cuda and compiler_name == CLANG:
                compilers.append((CLANG_CUDA, version))

    return compilers


@typechecked
def get_backend_matrix() -> List[List[Tuple[str, str]]]:
    """Generate backend list, where only backend is active on the same time.

    Returns:
        List[List[Tuple[str, str]]]: The backend list.
    """
    combination_matrix: List[List[Tuple[str, str]]] = []

    # TODO(SimeonEhrig) only working for HIP in the moment
    if HIPCC in sw_versions:
        for rocm_version in sw_versions[HIPCC]:
            combination_matrix.append([(ALPAKA_ACC_GPU_HIP_ENABLE, rocm_version)])

    if NVCC in sw_versions:
        for cuda_version in sw_versions[NVCC]:
            combination_matrix.append([(ALPAKA_ACC_GPU_CUDA_ENABLE, cuda_version)])

    return combination_matrix


@typechecked
def get_sw_tuple_list(name: str) -> List[Tuple[str, str]]:
    """Creates a list of software name version tuples for a software name.

    Args:
        name (str): Name of the software

    Returns:
        List[Tuple[str, str]]: List of software name versions tuples.
    """
    tuple_list: List[Tuple[str, str]] = []
    for version in sw_versions[name]:
        tuple_list.append((name, version))

    return tuple_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from conans import ConanFile, CMake
import os


class TestPackageConan(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
    generators = "cmake_find_package_multi", "cmake"

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def test(self):
        assert os.path.isfile(os.path.join(
            self.deps_cpp_info["catch2"].rootpath, "licenses", "LICENSE.txt"))
        bin_path = os.path.join("bin", "test_package")
        self.run("%s -s" % bin_path, run_environment=True)



"""Copyright 2023 Simeon Ehrig

This file is part of alpaka.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.


Verification of the results.
"""

from typing import List, Dict, Tuple
from typeguard import typechecked

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_job_coverage.util import strict_equal
import versions


@typechecked
def verify(combinations: List[Dict[str, Tuple[str, str]]]) -> bool:
    """Check if job matrix fullfill certain requirements.
    Args:
        combinations (List[Dict[str, Tuple[str, str]]]): The job matrix.

    Returns:
        bool: True if all checks passes, otherwise False.
    """

    # print("\033[31mverification failed\033[m")
    print("\033[33mWARNING: no verification tests implemented\033[m")
    # print("\033[32mverification was fine\033[m")

    return True



"""Copyright 2023 Simeon Ehrig

This file is part of alpaka.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.


Create GitLab-CI job description written in yaml from the job matrix."""

from typing import List, Dict, Tuple
from typeguard import typechecked
import os, yaml
from packaging import version as pk_version


from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_globals import *  # pylint: disable=wildcard-import,unused-wildcard-import

JOB_COMPILE_ONLY = "compile_only_job"
JOB_RUNTIME = "runtime_job"
JOB_ROCM_RUNTIME = "rocm_runtime_job"
JOB_NVCC_GCC_RUNTIME = "nvcc_gcc_runtime_job"
JOB_NVCC_CLANG_RUNTIME = "nvcc_clng_runtime_job"
JOB_CLANG_CUDA_RUNTIME = "clang_cuda_runtime_job"
JOB_UNKNOWN = "unknowm_job_type"

WAVE_GROUP_NAMES = [
    JOB_COMPILE_ONLY,
    JOB_RUNTIME,
    # can be enabled again, if fine granular scheduling is required
    # JOB_CPU_RUNTIME,
    # JOB_ROCM_RUNTIME,
    # JOB_NVCC_GCC_RUNTIME,
    # JOB_NVCC_CLANG_RUNTIME,
    # JOB_CLANG_CUDA_RUNTIME,
    JOB_UNKNOWN,
]


@typechecked
def get_env_var_name(variable_name: str) -> str:
    """Transform string to a shape, which is allowed as environment variable.

    Args:
        variable_name (str): Variable name.

    Returns:
        str: Transformed variable name.
    """
    return variable_name.upper().replace("-", "_")


@typechecked
def job_prefix_coding(job: Dict[str, Tuple[str, str]]) -> str:
    """Generate prefix for job name, depending of the available software versions.

    Args:
        job (Dict[str, Tuple[str, str]]): Job dict.

    Returns:
        str: Job name Prefix.
    """
    version_str = ""
    for sw in [CMAKE, BOOST, UBUNTU, CXX_STANDARD, BUILD_TYPE]:
        if sw in job:
            if job[sw][NAME] == CXX_STANDARD:
                version_str += "_cxx" + job[sw][VERSION]
            elif job[sw][NAME] == BUILD_TYPE:
                version_str += "_" + job[sw][VERSION]
            else:
                version_str += "_" + job[sw][NAME] + job[sw][VERSION]

    return version_str


@typechecked
def job_image(job: Dict[str, Tuple[str, str]], container_version: float) -> str:
    """Generate the image url deppending on the host and device compiler and the selected backend
    of the job.

    Args:
        job (Dict[str, Tuple[str, str]]): Job dict.
        container_version (float): Container version tag.

    Returns:
        str: Full container url, which can be used with docker pull.
    """
    container_url = "registry.hzdr.de/crp/alpaka-group-container/"
    container_url += "alpaka-ci-ubuntu" + job[UBUNTU][VERSION]

    # If only the GCC is used, use special gcc version of the container.
    if job[HOST_COMPILER][NAME] == GCC and job[DEVICE_COMPILER][NAME] == GCC:
        container_url += "-gcc"

    if (
        ALPAKA_ACC_GPU_CUDA_ENABLE in job
        and job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION] != OFF
        # TODO(SimeonEhrig) workaround until container version 3.1
        # container version 3.1 allows to ask, which images are available
        and pk_version.parse(job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION])
        < pk_version.parse("11.7")
    ):
        # Cast cuda version shape. E.g. from 11.0 to 110
        container_url += "-cuda" + str(
            int(float(job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION]) * 10)
        )
        if job[HOST_COMPILER][NAME] == GCC:
            container_url += "-gcc"

    if (
        ALPAKA_ACC_GPU_HIP_ENABLE in job
        and job[ALPAKA_ACC_GPU_HIP_ENABLE][VERSION] != OFF
    ):
        container_url += "-rocm" + job[ALPAKA_ACC_GPU_HIP_ENABLE][VERSION]

    # append container tag
    container_url += ":" + str(container_version)
    return container_url


@typechecked
def job_variables(job: Dict[str, Tuple[str, str]]) -> Dict[str, str]:
    """Add variables to the job depending of the job dict.

    Args:
        job (Dict[str, Tuple[str, str]]): Job dict

    Returns:
        Dict[str, str]: Dict of {variable name : variable value}.
    """
    variables: Dict[str, str] = {}

    ################################################################################################
    ### job independent environment variables
    ################################################################################################

    # OS name
    # at the moment, the GitLab CI has only Linux Runner, therefore it is fine to hard code it
    variables["ALPAKA_CI_OS_NAME"] = "Linux"
    variables["ALPAKA_CI_BUILD_JOBS"] = "$CI_CPUS"
    variables["OMP_NUM_THREADS"] = "$CI_CPUS"
    # the variable is required, that the test scripts, on which CI they are running
    variables["alpaka_CI"] = "GITLAB"
    variables["ALPAKA_CI_ANALYSIS"] = "OFF"
    variables["ALPAKA_CI_SANITIZERS"] = ""
    # cmake install path, if cmake version is not already installed
    variables["ALPAKA_CI_CMAKE_DIR"] = "$HOME/cmake"
    # boost install path, if boost version is not already installed
    variables["BOOST_ROOT"] = "$HOME/boost"
    # required, if boost is compiled during test job
    variables["ALPAKA_CI_BOOST_LIB_DIR"] = "$HOME/boost_libs"
    variables["BOOST_LIBRARYDIR"] = "/opt/boost/${ALPAKA_BOOST_VERSION}/lib"
    # cuda install path, if cuda version is not already installed
    variables["ALPAKA_CI_CUDA_DIR"] = "$HOME/cuda"
    # hip install path, if hip version is not already installed
    variables["ALPAKA_CI_HIP_ROOT_DIR"] = "$HOME/hip"

    ################################################################################################
    ### job dependent environment variables
    ################################################################################################

    variables["CMAKE_BUILD_TYPE"] = job[BUILD_TYPE][VERSION]
    if job[JOB_EXECUTION_TYPE][VERSION] == JOB_EXECUTION_RUNTIME:
        variables["ALPAKA_CI_RUN_TESTS"] = "ON"
    else:
        variables["ALPAKA_CI_RUN_TESTS"] = "OFF"

    variables["ALPAKA_CI_CMAKE_VER"] = job[CMAKE][VERSION]
    variables["ALPAKA_BOOST_VERSION"] = job[BOOST][VERSION]

    # TODO(SimeonEhrig) at the moment, not all backends are available in
    # versions.sw_versions[BACKENDS], therefore we have to set each backend explicit
    # Later, we can iterate over versions.sw_versions[BACKENDS] and check, if a Backend needs to be
    # enabled
    variables["alpaka_ACC_ANY_BT_OACC_ENABLE"] = "OFF"
    variables[ALPAKA_ACC_ANY_BT_OMP5_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_FIBERS_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE] = "ON"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE] = "OFF"
    variables[ALPAKA_ACC_GPU_CUDA_ENABLE] = "OFF"
    variables["alpaka_ACC_GPU_CUDA_ONLY_MODE"] = "OFF"
    variables[ALPAKA_ACC_GPU_HIP_ENABLE] = "OFF"
    variables["alpaka_ACC_GPU_HIP_ONLY_MODE"] = "OFF"

    if job[DEVICE_COMPILER][NAME] == HIPCC:
        variables[ALPAKA_ACC_GPU_HIP_ENABLE] = "ON"
        variables["CC"] = "clang"
        variables["CXX"] = "clang++"
        variables["GPU_TARGETS"] = "${CI_GPU_ARCH}"
        # TODO(SimeonEhrig) check, if we can remove this variable:
        if job[DEVICE_COMPILER][VERSION] == "5.0":
            variables["ALPAKA_CI_CLANG_VER"] = "14"
        elif job[DEVICE_COMPILER][VERSION] == "5.1":
            variables["ALPAKA_CI_CLANG_VER"] = "14"
        elif job[DEVICE_COMPILER][VERSION] == "5.2":
            variables["ALPAKA_CI_CLANG_VER"] = "14"
        elif job[DEVICE_COMPILER][VERSION] == "5.3":
            variables["ALPAKA_CI_CLANG_VER"] = "15"
        else:
            raise RuntimeError(
                "generate_job_yaml.job_variables(): unknown hip version: "
                f"{job[DEVICE_COMPILER][VERSION]}"
            )
        variables["ALPAKA_CI_HIP_VERSION"] = job[DEVICE_COMPILER][VERSION]
        variables["ALPAKA_CI_STDLIB"] = "libstdc++"

    # general configuration, if the CUDA backend is enabled (includes nvcc and clang as CUDA
    # compiler)
    if (
        ALPAKA_ACC_GPU_CUDA_ENABLE in job
        and job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION] != OFF
    ):
        variables[ALPAKA_ACC_GPU_CUDA_ENABLE] = "ON"
        variables["ALPAKA_CI_STDLIB"] = "libstdc++"
        variables["CMAKE_CUDA_ARCHITECTURES"] = job[SM_LEVEL][VERSION]
        variables["ALPAKA_CI_CUDA_VERSION"] = job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION]

    if job[DEVICE_COMPILER][NAME] == NVCC:
        # general configuration, if nvcc is the CUDA compiler
        variables["CMAKE_CUDA_COMPILER"] = "nvcc"

        # configuration, if GCC is the CUDA host compiler
        if job[HOST_COMPILER][NAME] == GCC:
            variables["CC"] = "gcc"
            variables["CXX"] = "g++"
            variables["ALPAKA_CI_GCC_VER"] = job[HOST_COMPILER][VERSION]
            variables[ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE] = "ON"
            variables[ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE] = "ON"
            variables[ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE] = "ON"
        # configuration, if Clang is the CUDA host compiler
        elif job[HOST_COMPILER][NAME] == CLANG:
            variables["CC"] = "clang"
            variables["CXX"] = "clang++"
            variables["ALPAKA_CI_CLANG_VER"] = job[HOST_COMPILER][VERSION]
            variables[ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE] = "ON"
            variables[ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE] = "ON"
        else:
            raise RuntimeError(
                "generate_job_yaml.job_variables(): unknown CUDA host compiler: "
                f"{job[HOST_COMPILER][NAME]}"
            )

    if job[DEVICE_COMPILER][NAME] == CLANG_CUDA:
        variables["CC"] = "clang"
        variables["CXX"] = "clang++"
        variables["ALPAKA_CI_CLANG_VER"] = job[DEVICE_COMPILER][VERSION]
        variables["CMAKE_CUDA_COMPILER"] = "clang++"
        variables[ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE] = "ON"
        variables[ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE] = "OFF"
        variables[ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE] = "OFF"

    return variables


@typechecked
def job_tags(job: Dict[str, Tuple[str, str]]) -> List[str]:
    """Add tags to select the correct runner, e.g. CPU only or Nvidia GPU.

    Args:
        job (Dict[str, Tuple[str, str]]): Job dict.

    Returns:
        List[str]: List of tags.
    """
    if job[JOB_EXECUTION_TYPE][VERSION] == JOB_EXECUTION_COMPILE_ONLY:
        return ["x86_64", "cpuonly"]

    if (
        ALPAKA_ACC_GPU_CUDA_ENABLE in job
        and job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION] != OFF
    ):
        return ["x86_64", "cuda"]
    if (
        ALPAKA_ACC_GPU_HIP_ENABLE in job
        and job[ALPAKA_ACC_GPU_HIP_ENABLE][VERSION] != OFF
    ):
        return ["x86_64", "rocm"]

    # fallback
    return ["x86_64", "cpuonly"]


def global_variables() -> Dict[str, str]:
    """Generate global variables for the test jobs.

    Returns:
        Dict[str, str]: global variables
    """
    variables: Dict[str, str] = {}

    variables["ALPAKA_CI_OS_NAME"] = "Linux"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_SEQ_ENABLE] = "ON"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_THREADS_ENABLE] = "ON"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_FIBERS_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_TBB_T_SEQ_ENABLE] = "OFF"
    variables[ALPAKA_ACC_CPU_B_OMP2_T_SEQ_ENABLE] = "ON"
    variables[ALPAKA_ACC_CPU_B_SEQ_T_OMP2_ENABLE] = "ON"
    variables[ALPAKA_ACC_ANY_BT_OMP5_ENABLE] = "OFF"
    variables["alpaka_ACC_ANY_BT_OACC_ENABLE"] = "OFF"
    variables[ALPAKA_ACC_GPU_CUDA_ENABLE] = "OFF"
    variables["alpaka_ACC_GPU_CUDA_ONLY_MODE"] = "OFF"
    variables[ALPAKA_ACC_GPU_HIP_ENABLE] = "OFF"
    variables["alpaka_ACC_GPU_HIP_ONLY_MODE"] = "OFF"
    # If ALPAKA_CI_ANALYSIS is OFF compile and execute runtime tests else compile only.
    variables["ALPAKA_CI_ANALYSIS"] = "OFF"
    variables["ALPAKA_CI_RUN_TESTS"] = "ON"
    variables["alpaka_CI"] = "GITLAB"
    variables["ALPAKA_CI_SANITIZERS"] = ""
    # TODO(SimeonEhrig) should be set in the before_install.sh and could be removed
    variables["ALPAKA_CI_INSTALL_CUDA"] = "OFF"
    # TODO(SimeonEhrig) should be set in the before_install.sh and could be removed
    variables["ALPAKA_CI_INSTALL_HIP"] = "OFF"
    variables["ALPAKA_CI_CMAKE_DIR"] = "$HOME/cmake"
    variables["BOOST_ROOT"] = "$HOME/boost"
    variables["ALPAKA_CI_BOOST_LIB_DIR"] = "$HOME/boost_libs"
    variables["ALPAKA_CI_CUDA_DIR"] = "$HOME/cuda"
    variables["ALPAKA_CI_HIP_ROOT_DIR"] = "$HOME/hip"

    return variables


@typechecked
def create_job(
    job: Dict[str, Tuple[str, str]], container_version: float
) -> Dict[str, Dict]:
    """Create complete GitLab-CI yaml for a single job

    Args:
        job (Dict[str, Tuple[str, str]]): Job dict.
        stage_number (int): Number of the stage. Required for the stage attribute.
        container_version (float): Container version tag.

    Returns:
        Dict[str, Dict]: Job yaml.
    """

    # the job name starts with the device compiler
    job_name = "linux_" + job[DEVICE_COMPILER][NAME] + job[DEVICE_COMPILER][VERSION]
    # if the nvcc is the device compiler, add also the host compiler to the name
    if job[DEVICE_COMPILER][NAME] == NVCC:
        job_name = (
            job_name + "-" + job[HOST_COMPILER][NAME] + job[HOST_COMPILER][VERSION]
        )
    # if Clang-CUDA is the device compiler, add also the CUDA SDK version to the name
    if job[DEVICE_COMPILER][NAME] == CLANG_CUDA:
        job_name = job_name + "-cuda" + job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION]

    if job[JOB_EXECUTION_TYPE][VERSION] == JOB_EXECUTION_COMPILE_ONLY:
        job_name += "_compile_only"

    # we need a really explicit job naming, otherwise job names are not unique
    # if a job name is not unique, only one version of the job is executed
    job_name += job_prefix_coding(job)

    job_yaml: Dict = {}
    job_yaml["image"] = job_image(job, container_version)
    job_yaml["variables"] = job_variables(job)
    job_yaml["script"] = [
        "source ./script/gitlabci/print_env.sh",
        "source ./script/gitlab_ci_run.sh",
    ]
    job_yaml["tags"] = job_tags(job)
    job_yaml["interruptible"] = True

    return {job_name: job_yaml}


@typechecked
def generate_job_yaml_list(
    job_matrix: List[Dict[str, Tuple[str, str]]],
    container_version: float,
) -> List[Dict[str, Dict]]:
    """Generate the job yaml for each job in the job matrix.

    Args:
        job_matrix (List[List[Dict[str, Tuple[str, str]]]]): Job Matrix
        container_version (float): Container version tag.

    Returns:
        List[Dict[str, Dict]]: List of GitLab-CI jobs. The key of a dict entry
        is the job name and the value is the body.
    """
    job_matrix_yaml: Dict[str, Dict] = []
    for job in job_matrix:
        job_matrix_yaml.append(create_job(job, container_version))

    return job_matrix_yaml


@typechecked
def distribute_to_waves(
    job_matrix: List[Dict[str, Dict]], wave_size: Dict[str, int] = {}
) -> Dict[str, List[List[Dict[str, Dict]]]]:

    sorted_groups = {}

    for wave in WAVE_GROUP_NAMES:
        sorted_groups[wave] = []

    for job in job_matrix:
        job_name: str = next(iter(job))
        if "_compile_only_" in job_name:
            sorted_groups[JOB_COMPILE_ONLY].append(job)
        # Clang as C++ compiler without CUDA backend
        elif job_name.startswith("linux_clang") and not job_name.startswith(
            "linux_clang-cuda"
        ):
            # sorted_groups[JOB_CPU_RUNTIME].append(job)
            sorted_groups[JOB_COMPILE_ONLY].append(job)
        elif job_name.startswith("linux_hipcc"):
            # sorted_groups[JOB_ROCM_RUNTIME].append(job)
            sorted_groups[JOB_RUNTIME].append(job)
        elif job_name.startswith("linux_nvcc") and "gcc" in job_name:
            # sorted_groups[JOB_NVCC_GCC_RUNTIME].append(job)
            sorted_groups[JOB_RUNTIME].append(job)
        elif job_name.startswith("linux_nvcc") and "clang" in job_name:
            # sorted_groups[JOB_NVCC_CLANG_RUNTIME].append(job)
            sorted_groups[JOB_RUNTIME].append(job)
        elif job_name.startswith("linux_clang-cuda"):
            # sorted_groups[JOB_CLANG_CUDA_RUNTIME].append(job)
            sorted_groups[JOB_RUNTIME].append(job)
        elif job_name.startswith("linux_nvhpc"):
            sorted_groups[JOB_RUNTIME].append(job)
        else:
            sorted_groups[JOB_UNKNOWN].append(job)

    for wave in WAVE_GROUP_NAMES:
        if not wave in wave_size:
            wave_size[wave] = len(sorted_groups[wave])
        else:
            # if max_jobs is negative, set to 0
            if wave_size[wave] < 0:
                wave_size[wave] = 0
            # if max_jobs is greater than len(sorted_groups[wave]), crop it to len(sorted_groups[wave])
            elif wave_size[wave] > len(sorted_groups[wave]):
                wave_size[wave] = len(sorted_groups[wave])

    wave_matrix: Dict[str : List[List[Dict[str, Dict]]]] = {}

    for wave in WAVE_GROUP_NAMES:
        wave_matrix[wave] = []

        # handle special case: range() expect step size bigger than 1
        wave_size[wave] = 1 if wave_size[wave] == 0 else wave_size[wave]

        for i in range(0, len(sorted_groups[wave]), wave_size[wave]):
            wave_matrix[wave].append(sorted_groups[wave][i : i + wave_size[wave]])

    return wave_matrix


@typechecked
def write_job_yaml(
    job_matrix: Dict[str, List[List[Dict[str, Dict]]]],
    path: str,
):
    """Write GitLab-CI jobs to file.

    Args:
        job_matrix (List[List[Dict[str, Dict]]]): List of GitLab-CI jobs. The
        key of a dict entry is the job name and the value is the body.
        path (str): Path of the GitLab-CI yaml file.
    """
    with open(path, "w", encoding="utf-8") as output_file:
        # If there is no CI job, create a dummy job.
        # This can happen if the filter filters out all jobs.
        number_of_jobs = 0
        for wave_name in WAVE_GROUP_NAMES:
            number_of_jobs += len(job_matrix[wave_name])

        if number_of_jobs == 0:
            yaml.dump(
                {
                    "dummy-job": {
                        "image": "alpine:latest",
                        "interruptible": True,
                        "script": [
                            'echo "This is a dummy job so that the CI does not fail."'
                        ],
                    }
                },
                output_file,
            )
            return

        stages: Dict[str, List[str]] = {"stages": []}

        for wave_name in WAVE_GROUP_NAMES:
            # setup all stages
            for stage_number in range(len(job_matrix[wave_name])):
                stages["stages"].append(f"{wave_name}-stage{stage_number}")

        yaml.dump(stages, output_file)
        output_file.write("\n")
        # add global variables for all jobs
        yaml.dump({"variables": global_variables()}, output_file)
        output_file.write("\n")

        # TODO: remove me, when all cuda and hip custom jobs are generated by the job generator
        # The CUDA and HIP jobs inherent from a job template written in yaml
        script_path = os.path.abspath(__file__)
        with open(
            os.path.abspath(
                os.path.join(os.path.dirname(script_path), "../gitlabci/job_base.yml")
            ),
            "r",
            encoding="utf8",
        ) as file:
            job_base_yaml = yaml.load(file, yaml.loader.SafeLoader)
        yaml.dump(job_base_yaml, output_file)

        for wave_name in WAVE_GROUP_NAMES:
            # Writes each job separately to the file.
            # If all jobs would be collected first in dict, the order would be not guarantied.
            for stage_number, wave in enumerate(job_matrix[wave_name]):
                # Improve the readability of the generated job yaml
                output_file.write(
                    f"# <<<<<<<<<<<<< {wave_name}-stage{stage_number} >>>>>>>>>>>>>\n\n"
                )
                for job in wave:
                    # the first key is the name
                    job[list(job.keys())[0]][
                        "stage"
                    ] = f"{wave_name}-stage{stage_number}"

                    yaml.dump(job, output_file)
                    output_file.write("\n")



"""Copyright 2023 Simeon Ehrig

This file is part of alpaka.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.

Alpaka project specific filter rules.
"""

from typing import List

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_globals import *  # pylint: disable=wildcard-import,unused-wildcard-import

from alpaka_job_coverage.util import (
    row_check_name,
    row_check_version,
    row_check_backend_version,
)


def alpaka_post_filter(row: List) -> bool:
    # disable clang as host compiler for nvcc 11.3 until 11.5
    # https://github.com/alpaka-group/alpaka/issues/1625
    if row_check_name(row, DEVICE_COMPILER, "==", NVCC) and row_check_name(
        row, HOST_COMPILER, "==", CLANG
    ):
        if row_check_version(row, DEVICE_COMPILER, ">=", "11.3") and row_check_version(
            row, DEVICE_COMPILER, "<=", "11.5"
        ):
            return False

    # disable all clang versions older than 14 as CUDA Compiler
    # https://github.com/alpaka-group/alpaka/issues/1857
    if row_check_backend_version(
        row, ALPAKA_ACC_GPU_CUDA_ENABLE, "!=", OFF
    ) and row_check_name(row, DEVICE_COMPILER, "==", CLANG_CUDA):
        if row_check_version(row, DEVICE_COMPILER, "<", "14"):
            return False

        # a bug in CMAKE 3.18 avoids the correct usage of the variable CMAKE_CUDA_ARCHITECTURE
        if row_check_version(row, CMAKE, "<", "3.19"):
            return False

    return True



"""Copyright 2023 Simeon Ehrig

This file is part of alpaka.

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.


Contains alpaka specific function to manipulate the values of a generated job matrix.
"""

from typing import List, Dict, Tuple
from packaging import version

from alpaka_job_coverage.globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
from alpaka_globals import *  # pylint: disable=wildcard-import,unused-wildcard-import
import versions


def add_job_parameters(job_matrix: List[Dict[str, Tuple[str, str]]]):
    """Add additional job parameters, which are depend of other parameters and has nothing to do
    with combinations. For example, decide if a job is a compile only or runtime test job

    Args:
        job_matrix (List[Dict[str, Tuple[str, str]]]): Job matrix
    """
    for hip_version in versions.sw_versions[HIPCC]:
        for job in job_matrix:
            if (
                job[DEVICE_COMPILER][NAME] == HIPCC
                and job[DEVICE_COMPILER][VERSION] == hip_version
                and job[BUILD_TYPE][VERSION] == CMAKE_DEBUG
            ):
                job[JOB_EXECUTION_TYPE] = (JOB_EXECUTION_TYPE, JOB_EXECUTION_RUNTIME)
                break

    ##############################################
    ## Defining runtime test for the CUDA backend
    ##############################################
    # The geneal rule is, that the latest minor release of each CUDA major release is tested with
    # all host compiler (one job per host compiler version)

    # This is a helper dictionary to find the latest minor version of a CUDA SDK major, used by a
    # specific host compiler.
    # e.g. We have CUDA SDK versions from 11.0 to 11.8 and 12.0 and 12.1. GCC as host compiler
    # supports all SDK versions, Clang as host compiler only the 11 versions and Clang as CUDA
    # compiler only up to 11.5. So the result is (see CUDA_SDK_per_compiler later):
    # {"GCC" : ["11.8", "12.1"], "Clang" : ["11.8"], "Clang-CUDA" : ["11.5"]}

    # { compiler_name : {major : (minor, (version_string)}}
    latest_CUDA_SDK_minor_versions: Dict[str : Dict[int:(int, str)]] = {
        GCC: {},
        CLANG: {},
        CLANG_CUDA: {},
    }

    MINOR_VERSION = 0
    VERSION_STRING = 1

    for job in job_matrix:
        if (
            ALPAKA_ACC_GPU_CUDA_ENABLE in job
            and job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION] != OFF
        ):
            v = version.parse(job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION])
            if not v.major in latest_CUDA_SDK_minor_versions[job[HOST_COMPILER][NAME]]:
                latest_CUDA_SDK_minor_versions[job[HOST_COMPILER][NAME]][v.major] = (
                    v.minor,
                    job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION],
                )
            elif (
                latest_CUDA_SDK_minor_versions[job[HOST_COMPILER][NAME]][v.major][
                    MINOR_VERSION
                ]
                < v.minor
            ):
                latest_CUDA_SDK_minor_versions[job[HOST_COMPILER][NAME]][v.major] = (
                    v.minor,
                    job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION],
                )

    CUDA_SDK_per_compiler: Dict[str : List[str]] = {}

    for compiler_name, version_values in latest_CUDA_SDK_minor_versions.items():
        CUDA_SDK_per_compiler[compiler_name] = []
        for sdk in version_values.values():
            CUDA_SDK_per_compiler[compiler_name].append(sdk[VERSION_STRING])

    # this stores, if a job with a specific host compiler version is already tagged as runtime job
    used_host_compiler: Dict[str, List[str]] = {}

    for compiler_name in CUDA_SDK_per_compiler.keys():
        used_host_compiler[compiler_name] = []

    for job in job_matrix:
        for compiler_name, sdk_versions in CUDA_SDK_per_compiler.items():
            if (
                ALPAKA_ACC_GPU_CUDA_ENABLE in job
                and job[HOST_COMPILER][NAME] == compiler_name
            ):
                for sdk_version in sdk_versions:
                    if (
                        job[ALPAKA_ACC_GPU_CUDA_ENABLE][VERSION] == sdk_version
                        # needs to be a release build, otherwise there is the risk of running ot of
                        # GPU resources
                        and job[BUILD_TYPE][VERSION] == CMAKE_RELEASE
                        and not job[HOST_COMPILER][VERSION]
                        in used_host_compiler[job[HOST_COMPILER][NAME]]
                    ):
                        used_host_compiler[job[HOST_COMPILER][NAME]].append(
                            job[HOST_COMPILER][VERSION]
                        )
                        job[JOB_EXECUTION_TYPE] = (
                            JOB_EXECUTION_TYPE,
                            JOB_EXECUTION_RUNTIME,
                        )

    # test one job per nvcc version with SM level of the CI GPU and highest supported SM level
    missing_nvcc_versions = versions.sw_versions[NVCC][:]

    STANDARD_SM_LEVEL = "61"
    for job in job_matrix:
        if (
            job[DEVICE_COMPILER][NAME] == NVCC
            and job[DEVICE_COMPILER][VERSION] in missing_nvcc_versions
            and job[JOB_EXECUTION_TYPE][VERSION] == JOB_EXECUTION_COMPILE_ONLY
        ):
            if version.parse(job[DEVICE_COMPILER][VERSION]) < version.parse("11.1"):
                job[SM_LEVEL] = (SM_LEVEL, STANDARD_SM_LEVEL + ";80")
            elif version.parse(job[DEVICE_COMPILER][VERSION]) < version.parse("11.5"):
                job[SM_LEVEL] = (SM_LEVEL, STANDARD_SM_LEVEL + ";86")
            elif version.parse(job[DEVICE_COMPILER][VERSION]) < version.parse("11.8"):
                job[SM_LEVEL] = (SM_LEVEL, STANDARD_SM_LEVEL + ";87")
            else:
                job[SM_LEVEL] = (SM_LEVEL, STANDARD_SM_LEVEL + ";90")
            missing_nvcc_versions.remove(job[DEVICE_COMPILER][VERSION])
        elif (
            ALPAKA_ACC_GPU_CUDA_ENABLE in job
            and ALPAKA_ACC_GPU_CUDA_ENABLE[VERSION] != OFF
        ):
            job[SM_LEVEL] = (SM_LEVEL, STANDARD_SM_LEVEL)
        else:
            job[SM_LEVEL] = (SM_LEVEL, "")




import sys
from pathlib import Path
from setuptools import setup
import subprocess

__version__ = "2.2.6"

this_directory = Path(__file__).parent
long_description = (this_directory/'README.md').read_text()

cmake_command = ['cmake', '-B', 'build']
make_command = ['make', '-C', 'build']

try:
    # Execute the cmake command and print its output
    subprocess.check_call(cmake_command, stderr=subprocess.STDOUT)
    # Execute the make command and print its output
    subprocess.check_call(make_command, stderr=subprocess.STDOUT)
except subprocess.CalledProcessError as e:
    print(e.output)
    sys.exit(e.returncode)


setup(
    name="CLUEstering",
    version=__version__,
    author="Simone Balducci",
    author_email="simone.balducci00@gmail.com",
    description='''A library that generalizes the original 2-dimensional CLUE
                   algorithm made at CERN.''',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['CLUEstering'],
    install_requires=['scikit-learn', 'numpy', 'matplotlib', 'pandas'],
    package_data={'': []},
    keywords=['Python', 'Clustering', 'Binding'],
    python_requires='>=3.7',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ]
)
